package com.vechain.vechainnfcverifydemo;

import android.Manifest;
import android.nfc.NfcAdapter;
import android.os.Bundle;
import android.util.Log;
import android.widget.Toast;

import com.tbruyelle.rxpermissions2.RxPermissions;
import com.vechain.common.ResultCode;
import com.vechain.prosdk.network.bean.VerifyResult;
import com.vechain.prosdk.nfctools.VeChainNFCSDK;
import com.victor.loading.rotate.RotateLoading;

import androidx.appcompat.app.AppCompatActivity;
import io.reactivex.disposables.Disposable;
import io.reactivex.functions.Consumer;

/***
 *
 * Simple example how to use VeChain Verify SDK libraries to read VeChain chips on android phone with NFC.
 *
 * This example show:
 * 1. Setup permissions to access NFC reader (in AndroidManifest.xml and check&request permissions on MainActivity)
 * 2. Setup scan card activity layout (MainActivity) to show user how to tap card according to the phone's NFC reader location
 */
public class MainActivity extends AppCompatActivity {

    private RxPermissions rxPermissions;
    private RotateLoading rotateLoading;

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_main);

        rotateLoading = findViewById(R.id.rotateloading);
        rxPermissions = new RxPermissions(this);

        /*
          Register SDK, Call this method before you start to verify.
         */

        VeChainNFCSDK.registerSDK(this, GlobalConfig.VerifyId, new MySignAction());

        String version = VeChainNFCSDK.getVersion();
        Log.d("VeChain", "onCreate: SDK Version " + version);
    }

    @Override
    protected void onResume() {
        super.onResume();

        checkSupport();
    }

    private void checkSupport() {
        //Check for NFC availability first
        NfcAdapter nfcAdapter = NfcAdapter.getDefaultAdapter(this.getApplicationContext());

        if (nfcAdapter == null) {
            //FEATURE_NFC is unavailable.
            return;
        }

        if (!nfcAdapter.isEnabled()) {
            //NFC is unable.
            return;
        }

        //Request NFC permission
        Disposable subscribe = rxPermissions.request(Manifest.permission.NFC)
                .subscribe(new Consumer<Boolean>() {
                    @Override
                    public void accept(Boolean isEnable) {
                        if (isEnable) {
                            enableNFCVerify();
                        }
                    }
                });

    }

    private void enableNFCVerify() {
        VeChainNFCSDK.onResume(MainActivity.this, new VeChainNFCSDK.NfcSDKCallback() {
            @Override
            public void onVerifyStart() {
                //a callback to be invoked when the system finds a tag while the foreground activity is operating in reader mode.
                //You can refresh UI (like a loading dialog) in this callback.
                boolean start = rotateLoading.isStart();
                if (!start) {
                    rotateLoading.start();
                }
            }

            @Override
            public void onResultReceived(VerifyResult verifyResult) {

                boolean start = rotateLoading.isStart();
                if (start) {
                    rotateLoading.stop();
                }

                int code = verifyResult.getCode();

                if (code == ResultCode.CODE_NETWORK_ERROR) {
                    //Network error
                    Toast.makeText(MainActivity.this, "Network Error", Toast.LENGTH_SHORT).show();
                    return;
                }

                if (code != ResultCode.OK) {
                    //verify fail
                    Toast.makeText(MainActivity.this, "Verify failed: error code is " + code, Toast.LENGTH_SHORT).show();
                    return;
                }

                //verify success
                String vid = verifyResult.getVid();

                // if tag contains
                String content = verifyResult.getContent();

                // vidAccessToken for get more data by vid
                String vidAccessToken = verifyResult.getVidAccessToken();

                Toast.makeText(MainActivity.this, "VeChain verified Vid is " + vid, Toast.LENGTH_SHORT).show();
            }
        });
    }

    @Override
    protected void onPause() {
        super.onPause();
        disableNFCVerify();
    }

    private void disableNFCVerify() {
        VeChainNFCSDK.onPause();
    }

    @Override
    protected void onDestroy() {
        super.onDestroy();
        VeChainNFCSDK.onDestroy();
    }
}
